import json
from types import MethodType
from typing import Any

from django.conf import settings
from django.db.models import Model
from django.utils.html import format_html

DELIMITER = "."


def encode_pk(pk: Any) -> str:
    return json.dumps(pk)


def decode_pk(encoded_pk: str) -> Any:
    return json.loads(encoded_pk)


def instance_ref_to_channel_group_key(app_label: str, model: str, pk: Any) -> str:
    return ".".join([app_label, model, str(pk)])


def serialise_model_attr(instance: Model, attr: str) -> str:
    """
    Returns a string representation of the value of
    this instance.
    """
    value = getattr(instance, attr)
    return serialise_value(value=value)


def serialise_value(*, value: Any) -> str:
    """
    Returns a string representation of the value of
    this instance.
    """
    if isinstance(value, Model):
        return str(value.pk)
    return str(value)


def set_original_form_values_on_instance(form, *, instance=None):
    """
    If the form has fields, look for the special __original-*
    fields. If provided, attach them to the associated instance,
    so they are respected when clean() is invoked
    """
    if not getattr(form, "fields", None):
        # form does not (yet) have fields, so nothing to do
        return

    for field_name, field in form.fields.items():
        field_key = (
            f"__original-{form.prefix}-{field_name}"
            if form.prefix
            else f"__original-{field_name}"
        )
        if field_key in form.data:
            original_value = form.data[field_key]
            (instance or form.instance)._original_form_values[
                field_name
            ] = original_value


def patch_widgets(form):
    """
    Modify the widgets so they will display the hidden inputs alongside
    the normal stuff.
    """
    websocket_connection_delay = getattr(
        settings, "NANGO_WEBSOCKET_CONNECTION_DELAY", -1
    )
    assert hasattr(form, "fields")
    for field_name, field in form.fields.items():
        widget = field.widget

        def _render(
            self,
            template,
            context,
            *args,
            _original_render=widget._render,
            _field_name=field_name,
            **kw,
        ):
            result = _original_render(template, context, *args, **kw)
            if "widget" not in context:
                return result

            hidden = format_html(
                f"""<input data-app-label="{form.instance._meta.app_label}"
                               data-model-name="{form.instance._meta.model_name}"
                               data-instance-pk="{encode_pk(form.instance.pk)}"
                               data-original-name="{{}}"
                               data-related-form-id="{{}}"
                               type="hidden"
                               data-connection-delay="{websocket_connection_delay}"
                               name="__original-{{}}"
                               value="{{}}">""",
                _field_name,
                context["widget"]["attrs"]["id"],
                context["widget"]["name"],
                serialise_model_attr(form.instance, _field_name),
            )
            return result + hidden

        widget._render = MethodType(_render, widget)



from functools import partial
from inspect import isclass
from inspect import signature
from typing import Any

from django import forms
from nango.common import patch_widgets
from nango.common import set_original_form_values_on_instance


class FormMixin:
    def __init__(self, *args, **kw) -> None:
        super().__init__(*args, **kw)
        if fields := getattr(self, "fields", None):
            patch_widgets(form=self)

    def clean(self, *args, **kw):
        set_original_form_values_on_instance(form=self)
        return super().clean(*args, **kw)


class ModelForm(FormMixin, forms.ModelForm):
    pass


class Form(FormMixin, forms.Form):
    pass


def __getattr__(name: str) -> Any:
    original = getattr(forms, name)
    if isclass(original):
        # Anything derived from django's ModelForm gets
        # modified to have FormMixin too
        if issubclass(original, forms.Form) and not issubclass(original, FormMixin):

            class NewForm(FormMixin, original):  # type: ignore
                pass

            return NewForm
        return original

    if callable(original):
        sig = signature(original)
        # Anything with a 'form' argument which is derived from
        # django's ModelForm gets modified to have FormMixin too
        if form := sig.parameters.get("form"):
            default = form.default
            if issubclass(default, forms.Form) and not issubclass(default, FormMixin):

                class NewForm(FormMixin, default):  # type: ignore
                    pass

            return partial(original, form=NewForm)
        return original
    raise AttributeError(f"module {__name__} has no attribute {name}")



import json
import logging
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import Set
from typing import Tuple

from asgiref.sync import sync_to_async
from channels.generic.websocket import AsyncWebsocketConsumer
from django.apps import apps

from .common import decode_pk
from .common import instance_ref_to_channel_group_key
from .common import serialise_value

LOGGER = logging.getLogger(__file__)

MessageType = Tuple[str, str, Any]


class LiveUpdatesConsumer(AsyncWebsocketConsumer):  # type: ignore
    def __init__(self, **kw):
        super().__init__(**kw)
        self._my_groups: Set[str] = set()
        self.fields_for_instance: Dict[MessageType, Dict[str, Any]] = defaultdict(dict)

    async def connect(self) -> None:
        LOGGER.info("connected")
        await self.accept()

    def _retrieve_instance_values(self, app, model, pk, attrs):
        Model = apps.get_model(app, model)
        instance = Model.objects.get(pk=pk)
        return {attr: getattr(instance, attr) for attr in attrs}

    async def saved(self, info) -> None:
        print("saved")
        message = info["message"]
        app = message["app"]
        model = message["model"]
        pkey = message["pk"]
        fields = self.fields_for_instance[(app, model, pkey)]
        instance = await sync_to_async(self._retrieve_instance_values)(
            app=app, model=model, pk=pkey, attrs=fields
        )
        print(f"{fields=} {instance=}")
        for attr, original_value in fields.items():
            new_value = serialise_value(value=instance[attr])
            print(f"{attr=} {original_value=} {new_value=}")
            if original_value != new_value:
                message = dict(
                    message=dict(
                        app=app,
                        model=model,
                        pk=pkey,
                        attr=attr,
                        original_value=original_value,
                        new_value=new_value,
                    )
                )
                print(message)
                await self.send(text_data=json.dumps(message))
                fields[attr] = new_value

    async def receive(self, text_data: bytes) -> None:
        text_data_json = json.loads(text_data)
        LOGGER.error(f"{text_data_json=!r}")
        for instance_ref in text_data_json["register"]:
            app_label = instance_ref["appLabel"]
            model = instance_ref["model"]
            pkey = decode_pk(instance_ref["pk"])
            group_key = instance_ref_to_channel_group_key(
                app_label=app_label, model=model, pk=pkey
            )
            fields = self.fields_for_instance[(app_label, model, pkey)]
            fields[instance_ref["attr"]] = instance_ref["value"]
            self._my_groups.add(group_key)

        for group_key in self._my_groups:
            await self.channel_layer.group_add(
                channel=self.channel_name, group=group_key
            )
            await self.saved(
                info=dict(message=dict(app=app_label, model=model, pk=pkey))
            )
            print(f"{self._my_groups=}")

    async def disconnect(self, close_code: Any) -> None:
        LOGGER.info("disconnected")
        for my_group in self._my_groups:
            await self.channel_layer.group_discard(
                group=my_group, channel=self.channel_name
            )



import logging

from django.conf import settings
from django.urls import re_path

from . import consumers

LOGGER = logging.getLogger(__file__)

websocket_urlpatterns = []

if getattr(settings, "NANGO_WEBSOCKET_CONNECTION_DELAY", -1) >= 0:
    websocket_urlpatterns.append(
        re_path(r"ws/liveupdates/$", consumers.LiveUpdatesConsumer.as_asgi())
    )
else:
    LOGGER.warning("Nango websockets are disabled as connection delay < 0")



from inspect import isclass
from typing import Any

from django import forms
from django.contrib import admin
from nango import forms
from nango.common import set_original_form_values_on_instance


class AdminMixin:
    def save_model(self, request, obj, form, change):
        set_original_form_values_on_instance(form=form, instance=obj)
        return super().save_model(request=request, obj=obj, form=form, change=change)


def __getattr__(name: str) -> Any:
    OriginalClass = getattr(admin, name)
    if not isclass(OriginalClass):
        raise AttributeError(f"module {__name__} has no attribute {name}")

    """
    if issubclass(OriginalClass, forms.ModelForm):

        class NewForm(forms.FormMixin, OriginalClass):  # type: ignore
            pass

        return NewForm
    """

    Form = OriginalClass.form
    if not issubclass(Form, forms.FormMixin):

        class NewForm(forms.FormMixin, Form):  # type: ignore
            pass

        Form = NewForm

    class NewAdmin(AdminMixin, OriginalClass):  # type: ignore
        form = Form

    return NewAdmin


class ModelAdmin(AdminMixin, admin.ModelAdmin):
    form = forms.ModelForm
